package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.client.screen.panels.DockProperty

@Suppress("unused")
data class SkinGrid(
	val texture: ResourceLocation,
	val width: Float,
	val height: Float,
	val rows: Int = 16,
	val columns: Int = 16,
) {
	var row = 0
	var column = 0

	val imageWidth get() = width * columns
	val imageHeight get() = height * rows

	val currentX: Float get() = column * width
	val currentY: Float get() = row * height

	fun skip(): SkinGrid {
		column++

		if (column >= columns) {
			jump()
		}

		return this
	}

	fun jump(): SkinGrid {
		row++
		column = 0
		return this
	}

	fun next(): SkinElement {
		val element = SkinElement(texture, currentX, currentY, width, height, imageWidth, imageHeight)
		skip()
		return element
	}

	fun reset(): SkinGrid {
		row = 0
		column = 0
		return this
	}

	fun seek(row: Int, column: Int): SkinGrid {
		this.row = row
		this.column = column
		return this
	}

	operator fun get(column: Int, row: Int) = SkinElement(texture, column * width, row * height, width, height, imageWidth, imageHeight)
}

fun ResourceLocation.element(
	x: Float,
	y: Float,
	width: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, height, textureWidth, textureHeight)

fun ResourceLocation.pixel(
	x: Float,
	y: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, 1f, textureWidth, textureHeight)

fun ResourceLocation.hLine(
	x: Float,
	y: Float,
	width: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, 1f, textureWidth, textureHeight)

fun ResourceLocation.vLine(
	x: Float,
	y: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, height, textureWidth, textureHeight)

@Suppress("unused")
data class SkinElement @JvmOverloads constructor(
	val texture: ResourceLocation,
	val x: Float,
	val y: Float,
	val width: Float,
	val height: Float,
	val imageWidth: Float = 256f,
	val imageHeight: Float = 256f,
	val winding: UVWindingOrder = UVWindingOrder.NORMAL
) {
	init {
		require(x >= 0f) { "Invalid x $x" }
		require(y >= 0f) { "Invalid y $y" }
		require(width > 0f) { "Invalid width $width" }
		require(height > 0f) { "Invalid height $height" }
		require(imageWidth > 0f) { "Invalid image width $imageWidth" }
		require(imageHeight > 0f) { "Invalid image height $imageHeight" }
	}

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRaw(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = render(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRawPartial(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderWidth(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		winding: UVWindingOrder = this.winding
	) {
		render(stack, x, y, width = width, winding = winding)
	}

	@JvmOverloads
	fun renderWidth(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		winding: UVWindingOrder = this.winding,
	) = renderWidth(stack, x.toFloat(), y.toFloat(), width.toFloat(), winding)

	@JvmOverloads
	fun renderHieght(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding,
	) {
		render(stack, x, y, height = height, winding = winding)
	}

	@JvmOverloads
	fun renderHieght(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding,
	) = renderHieght(stack, x.toFloat(), y.toFloat(), height.toFloat(), winding)

	private val u0 = this.x / imageWidth
	private val v0 = this.y / imageHeight
	private val u1 = (this.x + width) / imageWidth
	private val v1 = (this.y + height) / imageHeight

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = UVWindingOrder.NORMAL
	) {
		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.NORMAL
	) = renderRaw(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = UVWindingOrder.NORMAL
	) {
		val u1 = (this.x + width) / imageWidth
		val v1 = (this.y + height) / imageHeight

		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.NORMAL
	) = renderRawPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)
}

data class StretchingRectangleElement(
	val topLeft: SkinElement,
	val topRight: SkinElement,
	val bottomLeft: SkinElement,
	val bottomRight: SkinElement,
	val left: SkinElement,
	val right: SkinElement,
	val top: SkinElement,
	val bottom: SkinElement,
	val middle: SkinElement,
	val padding: DockProperty = DockProperty.EMPTY
) {
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float,
		height: Float,
	) {
		val bgWidth = width - topLeft.width.coerceAtMost(bottomLeft.width) - topRight.width.coerceAtMost(bottomRight.width) - padding.left - padding.right
		val bgHeight = height - topLeft.height.coerceAtMost(bottomLeft.height) - topRight.height.coerceAtMost(bottomRight.height) - padding.top - padding.bottom

		if (bgWidth > 0 && bgHeight > 0) {
			middle.render(stack, x + topLeft.width + padding.left, y + topLeft.height + padding.top, bgWidth, bgHeight)
		}

		top.render(stack, x + topLeft.width, y, width = width - topLeft.width - topRight.width)
		bottom.render(stack, x + bottomLeft.width, y + height - bottom.height, width = width - bottomLeft.width - bottomRight.width)

		left.render(stack, x, y + topLeft.height, height = height - topLeft.height - bottomLeft.height)
		right.render(stack, x + width - right.width, y + topRight.height, height = height - topRight.height - bottomRight.height)

		topLeft.render(stack, x, y)
		topRight.render(stack, x + width - topRight.width, y)

		bottomLeft.render(stack, x, y + height - bottomLeft.height)
		bottomRight.render(stack, x + width - bottomRight.width, y + height - bottomLeft.height)
	}
}

